
##__________________________________________________________________||
QUERY_GLOBAL_STATE = '''
query GlobalState {
  globalState
}
'''.strip()

QUERY_SOURCE = '''
query Source($fileName: String) {
  source(fileName: $fileName)
}
'''.strip()

QUERY_SOURCE_LINE = '''
query SourceLine(
  $lineNo: Int!
  $fileName: String
) {
  sourceLine(lineNo: $lineNo, fileName: $fileName)
}
'''.strip()

QUERY_EXCEPTION = '''
query Exception {
  exception
}
'''.strip()

##__________________________________________________________________||
SUBSCRIBE_GLOBAL_STATE = '''
subscription GlobalState {
  globalState
}
'''.strip()

SUBSCRIBE_THREAD_TASK_IDS = '''
subscription ThreadTaskIds {
  threadTaskIds {
    threadId
    taskId
  }
}
'''.strip()

SUBSCRIBE_THREAD_TASK_STATE = '''
subscription ThreadTaskState(
  $threadId: String!
  $taskId: String
) {
  threadTaskState(threadId: $threadId, taskId: $taskId) {
    prompting
    fileName
    lineNo
    traceEvent
  }
}
'''.strip()

##__________________________________________________________________||
MUTATE_EXEC = '''
mutation Exec {
  exec
}
'''.strip()


MUTATION_RESET = '''
mutation Reset($statement: String) {
  reset(statement: $statement)
}
'''.strip()

MUTATE_SEND_PDB_COMMAND = '''
mutation SendPdbCommand(
  $threadId: String!
  $taskId: String
  $command: String!
) {
  sendPdbCommand(threadId: $threadId, taskId: $taskId, command: $command)
}
'''.strip()




import datetime
from sqlalchemy.ext.declarative import declarative_base

from sqlalchemy import Column, Integer, String, DateTime


Base = declarative_base()


class Hello(Base):
    __tablename__ = "hello"
    id = Column(Integer, primary_key=True, index=True)
    message = Column(String)


class StateChange(Base):
    __tablename__ = "state_change_log"
    id = Column(Integer, primary_key=True, index=True)
    name = Column(String)
    time = Column(DateTime(), default=lambda: datetime.datetime.now())



import sys
from async_asgi_testclient import TestClient

import pytest
from unittest.mock import AsyncMock, Mock

from nextlinegraphql import create_app


##__________________________________________________________________||
@pytest.fixture(autouse=True)
def mock_asyncio_sleep(monkeypatch):
    import asyncio

    y = Mock(wraps=asyncio)
    y.sleep = AsyncMock()
    module = sys.modules["nextlinegraphql.schema.bindables"]
    monkeypatch.setattr(module, "asyncio", y)
    yield y


##__________________________________________________________________||
@pytest.mark.asyncio
async def test_query():

    query = """
      { hello }
    """
    data = {"query": query}

    headers = {
        "user-agent": "Mozilla/5.0",
        "Content-Type:": "application/json",
    }

    async with TestClient(create_app()) as client:
        resp = await client.post("/", json=data, headers=headers)
        assert resp.status_code == 200
        expect = {"data": {"hello": "Hello, Mozilla/5.0!"}}
        assert expect == resp.json()


##__________________________________________________________________||
@pytest.mark.asyncio
async def test_subscription():

    query = """
      subscription {
        counter
      }
    """

    data = {
        "id": "1",
        "type": "start",
        "payload": {
            "variables": {},
            "extensions": {},
            "operationName": None,
            "query": query,
        },
    }

    async with TestClient(create_app()) as client:
        async with client.websocket_connect("/") as ws:
            await ws.send_json(data)

            expect = {
                "type": "data",
                "id": "1",
                "payload": {"data": {"counter": 1}},
            }
            actual = await ws.receive_json()
            assert expect == actual

            expect = {
                "type": "data",
                "id": "1",
                "payload": {"data": {"counter": 2}},
            }
            actual = await ws.receive_json()
            assert expect == actual

            expect = {
                "type": "data",
                "id": "1",
                "payload": {"data": {"counter": 3}},
            }
            actual = await ws.receive_json()
            assert expect == actual

            expect = {
                "type": "data",
                "id": "1",
                "payload": {"data": {"counter": 4}},
            }
            actual = await ws.receive_json()
            assert expect == actual

            expect = {
                "type": "data",
                "id": "1",
                "payload": {"data": {"counter": 5}},
            }
            actual = await ws.receive_json()
            assert expect == actual

            expect = {"type": "complete", "id": "1"}
            actual = await ws.receive_json()
            assert expect == actual


##__________________________________________________________________||



import sys
import asyncio
import threading
import traceback
import janus
from ariadne import ScalarType, QueryType, MutationType, SubscriptionType

from ...nl import run_nextline, reset_nextline, close_nextline


##__________________________________________________________________||
datetime_scalar = ScalarType("Datetime")


@datetime_scalar.serializer
def serialize_datetime(value):
    print(value)
    return value.isoformat()


##__________________________________________________________________||
query = QueryType()


@query.field("hello")
async def resolve_hello(_, info):
    request = info.context["request"]
    user_agent = request.headers.get("user-agent", "guest")
    return "Hello, %s!" % user_agent


@query.field("helloDb")
async def resolve_hello_db(_, info):
    db = info.context["db"]
    with db.Session.begin() as session:
        model = session.query(db.models.Hello).one_or_none()
        if model is None:
            return None
        return model.message


@query.field("stateChanges")
async def resolve_state_changes(_, info):
    db = info.context["db"]
    with db.Session.begin() as session:
        models = session.query(db.models.StateChange).all()
        return [{"name": m.name, "time": m.time} for m in models]


@query.field("globalState")
async def resolve_global_state(_, info):
    nextline = info.context["nextline"]
    return nextline.global_state


@query.field("source")
async def resolve_source(_, info, fileName=None):
    nextline = info.context["nextline"]
    return nextline.get_source(fileName)


@query.field("sourceLine")
async def resolve_source_line(_, info, lineNo, fileName=None):
    nextline = info.context["nextline"]
    return nextline.get_source_line(lineNo, fileName)


@query.field("exception")
async def resolve_exception(_, info):
    nextline = info.context["nextline"]
    exc = nextline.exception()
    if not exc:
        return
    ret = "".join(
        traceback.format_exception(type(exc), exc, exc.__traceback__)
    )
    return ret


##__________________________________________________________________||
subscription = SubscriptionType()


@subscription.source("counter")
async def counter_generator(obj, info):
    for i in range(5):
        await asyncio.sleep(1)
        yield i


@subscription.field("counter")
def counter_resolver(count, info):
    return count + 1


@subscription.source("globalState")
async def global_state_generator(_, info):
    nextline = info.context["nextline"]
    async for s in nextline.subscribe_global_state():
        # if s == 'finished':
        #     print('finished', nextline.exception())
        #     nextline.result()
        yield s


@subscription.field("globalState")
def global_state_resolver(global_state, info):
    return global_state


@subscription.source("threadTaskIds")
async def thread_task_ids_generator(_, info):
    nextline = info.context["nextline"]
    async for y in nextline.subscribe_thread_asynctask_ids():
        yield [{"threadId": e[0], "taskId": e[1]} for e in y]


@subscription.field("threadTaskIds")
def thread_task_ids_resolver(obj, info):
    return obj


@subscription.source("threadTaskState")
async def thread_task_state_generator(_, info, threadId, taskId):
    threadId = int(threadId)
    taskId = int(taskId) if taskId else None
    thread_asynctask_id = (threadId, taskId)
    nextline = info.context["nextline"]
    async for y in nextline.subscribe_thread_asynctask_state(
        thread_asynctask_id
    ):
        # if y['prompting'] and y['trace_event'] == 'return':
        #     nextline.send_pdb_command(thread_asynctask_id, 'return')
        #     continue
        y = {
            "prompting": y["prompting"],
            "fileName": y["file_name"],
            "lineNo": y["line_no"],
            "traceEvent": y["trace_event"],
        }
        yield y


@subscription.field("threadTaskState")
def thread_task_state_resolver(obj, info, threadId, taskId):
    return obj


@subscription.source("stdout")
async def stdout_generator(_, info):
    stdout_queue = await get_stdout_queue()
    queue = stdout_queue.subscribe()
    nextline = info.context["nextline"]
    while True:
        v = await queue.async_q.get()
        if nextline.global_state == "running":
            yield v
    await stdout_queue.unsubscribe(queue)


@subscription.field("stdout")
def stdout_resolver(stdout, info):
    return stdout


##__________________________________________________________________||
mutation = MutationType()


@mutation.field("exec")
async def resolve_exec(_, info):
    # print(_)
    # print(info)
    await run_nextline()
    return True


@mutation.field("reset")
async def resolve_reset(_, info, statement=None):
    reset_nextline(statement=statement)
    return True


@mutation.field("close")
async def resolve_close(_, info):
    await close_nextline()
    return True


@mutation.field("sendPdbCommand")
async def resolve_send_pdb_command(_, info, threadId, taskId, command):
    threadId = int(threadId)
    taskId = int(taskId) if taskId else None
    thread_asynctask_id = (threadId, taskId)
    nextline = info.context["nextline"]
    nextline.send_pdb_command(thread_asynctask_id, command)
    return True


@mutation.field("updateHelloDbMessage")
async def resolve_update_hello_db_message(_, info, message):
    db = info.context["db"]
    with db.Session.begin() as session:
        model = session.query(db.models.Hello).one_or_none()
        if model is None:
            model = db.models.Hello(message=message)
            session.add(model)
        else:
            model.message = message
        session.commit()
    return True


##__________________________________________________________________||
class StreamOut:
    def __init__(self, queue):
        self.queue = queue

    def write(self, s):
        self.queue.put(s)

    def flush(self):
        pass


class QueueDist:
    def __init__(self, queue: janus.Queue):
        self.queue = queue
        self.subscribers = []

        self.t = threading.Thread(target=self._listen, daemon=True)
        self.t.start()

    def _listen(self):
        while True:
            v = self.queue.sync_q.get()
            for s in self.subscribers:
                s.sync_q.put(v)

    def subscribe(self):
        ret = janus.Queue()
        self.subscribers.append(ret)
        return ret

    async def unsubscribe(self, queue):
        self.subscribers.remove(queue)
        queue.close()
        await queue.wait_closed()


class StdoutQueue:
    def __init__(self, queue: janus.Queue):
        self.queue = queue
        self.stdout_org = sys.stdout
        sys.stdout = StreamOut(self.queue.sync_q)
        self.queue_dist = QueueDist(self.queue)

        self.q = self.subscribe()
        self.t = threading.Thread(target=self._listen, daemon=True)
        self.t.start()

    def _listen(self):
        while True:
            v = self.q.sync_q.get()
            self.stdout_org.write(v)

    def subscribe(self):
        return self.queue_dist.subscribe()

    async def unsubscribe(self, queue):
        await self.queue_dist.unsubscribe(queue)


stdout_queue_holder = []


async def get_stdout_queue():
    if not stdout_queue_holder:
        queue = janus.Queue()
        stdout_queue_holder.append(StdoutQueue(queue))
    return stdout_queue_holder[0]


##__________________________________________________________________||
bindables = [query, mutation, subscription, datetime_scalar]

##__________________________________________________________________||



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema 1'] = {
    'data': {
        '__schema': {
            'types': [
                {
                    'description': 'A GraphQL Schema defines the capabilities of a GraphQL server. It exposes all available types and directives on the server, as well as the entry points for query, mutation, and subscription operations.',
                    'fields': [
                        {
                            'description': None,
                            'name': 'description',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': 'A list of all types supported by this server.',
                            'name': 'types',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': 'The type that query operations will be rooted at.',
                            'name': 'queryType',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': 'If this server supports mutation, the type that mutation operations will be rooted at.',
                            'name': 'mutationType',
                            'type': {
                                'name': '__Type'
                            }
                        },
                        {
                            'description': 'If this server support subscription, the type that subscription operations will be rooted at.',
                            'name': 'subscriptionType',
                            'type': {
                                'name': '__Type'
                            }
                        },
                        {
                            'description': 'A list of all directives supported by this server.',
                            'name': 'directives',
                            'type': {
                                'name': None
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': '__Schema'
                },
                {
                    'description': 'The `String` scalar type represents textual data, represented as UTF-8 character sequences. The String type is most often used by GraphQL to represent free-form human-readable text.',
                    'fields': None,
                    'inputFields': None,
                    'name': 'String'
                },
                {
                    'description': '''The fundamental unit of any GraphQL Schema is the type. There are many kinds of types in GraphQL as represented by the `__TypeKind` enum.

Depending on the kind of a type, certain fields describe information about that type. Scalar types provide no information beyond a name, description and optional `specifiedByUrl`, while Enum types provide their values. Object and Interface types provide the fields they describe. Abstract types, Union and Interface, provide the Object types possible at runtime. List and NonNull types compose other types.''',
                    'fields': [
                        {
                            'description': None,
                            'name': 'kind',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'name',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'description',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'specifiedByUrl',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'fields',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'interfaces',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'possibleTypes',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'enumValues',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'inputFields',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'ofType',
                            'type': {
                                'name': '__Type'
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': '__Type'
                },
                {
                    'description': 'An enum describing what kind of type a given `__Type` is.',
                    'fields': None,
                    'inputFields': None,
                    'name': '__TypeKind'
                },
                {
                    'description': 'Object and Interface types are described by a list of Fields, each of which has a name, potentially a list of arguments, and a return type.',
                    'fields': [
                        {
                            'description': None,
                            'name': 'name',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'description',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'args',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'type',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'isDeprecated',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'deprecationReason',
                            'type': {
                                'name': 'String'
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': '__Field'
                },
                {
                    'description': 'Arguments provided to Fields or Directives and the input fields of an InputObject are represented as Input Values which describe their type and optionally a default value.',
                    'fields': [
                        {
                            'description': None,
                            'name': 'name',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'description',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'type',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': 'A GraphQL-formatted string representing the default value for this input value.',
                            'name': 'defaultValue',
                            'type': {
                                'name': 'String'
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': '__InputValue'
                },
                {
                    'description': 'The `Boolean` scalar type represents `true` or `false`.',
                    'fields': None,
                    'inputFields': None,
                    'name': 'Boolean'
                },
                {
                    'description': 'One possible value for a given Enum. Enum values are unique values, not a placeholder for a string or numeric value. However an Enum value is returned in a JSON response as a string.',
                    'fields': [
                        {
                            'description': None,
                            'name': 'name',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'description',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'isDeprecated',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'deprecationReason',
                            'type': {
                                'name': 'String'
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': '__EnumValue'
                },
                {
                    'description': '''A Directive provides a way to describe alternate runtime execution and type validation behavior in a GraphQL document.

In some cases, you need to provide options to alter GraphQL's execution behavior in ways field arguments will not suffice, such as conditionally including or skipping a field. Directives provide this by describing additional information to the executor.''',
                    'fields': [
                        {
                            'description': None,
                            'name': 'name',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'description',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'isRepeatable',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'locations',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'args',
                            'type': {
                                'name': None
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': '__Directive'
                },
                {
                    'description': 'A Directive can be adjacent to many parts of the GraphQL language, a __DirectiveLocation describes one such possible adjacencies.',
                    'fields': None,
                    'inputFields': None,
                    'name': '__DirectiveLocation'
                },
                {
                    'description': None,
                    'fields': [
                        {
                            'description': None,
                            'name': 'hello',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'helloDb',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'globalState',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'source',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'sourceLine',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'exception',
                            'type': {
                                'name': 'String'
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'Query'
                },
                {
                    'description': 'The `Int` scalar type represents non-fractional signed whole numeric values. Int can represent values between -(2^31) and 2^31 - 1.',
                    'fields': None,
                    'inputFields': None,
                    'name': 'Int'
                },
                {
                    'description': None,
                    'fields': [
                        {
                            'description': None,
                            'name': 'counter',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'globalState',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'threadTaskIds',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'threadTaskState',
                            'type': {
                                'name': 'ThreadTaskState'
                            }
                        },
                        {
                            'description': None,
                            'name': 'stdout',
                            'type': {
                                'name': 'String'
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'Subscription'
                },
                {
                    'description': None,
                    'fields': [
                        {
                            'description': None,
                            'name': 'exec',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'reset',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'close',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'sendPdbCommand',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'updateHelloDbMessage',
                            'type': {
                                'name': None
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'Mutation'
                },
                {
                    'description': None,
                    'fields': [
                        {
                            'description': None,
                            'name': 'threadId',
                            'type': {
                                'name': None
                            }
                        },
                        {
                            'description': None,
                            'name': 'taskId',
                            'type': {
                                'name': 'String'
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'ThreadTaskId'
                },
                {
                    'description': None,
                    'fields': [
                        {
                            'description': None,
                            'name': 'prompting',
                            'type': {
                                'name': 'Int'
                            }
                        },
                        {
                            'description': None,
                            'name': 'fileName',
                            'type': {
                                'name': 'String'
                            }
                        },
                        {
                            'description': None,
                            'name': 'lineNo',
                            'type': {
                                'name': 'Int'
                            }
                        },
                        {
                            'description': None,
                            'name': 'traceEvent',
                            'type': {
                                'name': 'String'
                            }
                        }
                    ],
                    'inputFields': None,
                    'name': 'ThreadTaskState'
                }
            ]
        }
    }
}



import asyncio
from ariadne.asgi import GraphQL

from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.middleware.cors import CORSMiddleware

from .schema import schema
from .db import Db
from .nl import get_nextline


##__________________________________________________________________||
class WGraphQL(GraphQL):
    """Wrap GraphQL for introspection"""

    async def __call__(self, scope, receive, send):
        # print(scope)
        # print(receive)
        # print(send)
        await super().__call__(scope, receive, send)

    async def handle_websocket_message(
        self, message, websocket, subscriptions
    ):
        # print(message)
        # print(websocket)
        # print(subscriptions)
        await super().handle_websocket_message(
            message, websocket, subscriptions
        )


async def monitor_state(db):
    nextline = get_nextline()
    async for s in nextline.subscribe_global_state():
        with db.Session.begin() as session:
            model = db.models.StateChange(name=s)
            session.add(model)
            session.commit()


def create_app():

    db = Db()

    asyncio.create_task(monitor_state(db))

    app_ = WGraphQL(
        schema,
        context_value=lambda request: {
            "request": request,
            "db": db,
            "nextline": get_nextline(),
        },
        debug=True,
    )

    middleware = [
        Middleware(
            CORSMiddleware,
            allow_origins=["*"],
            allow_methods=["*"],
            allow_headers=["*"],
        )
    ]

    app = Starlette(debug=True, middleware=middleware)
    app.mount("/", app_)

    return app


##__________________________________________________________________||



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_source[default] 1'] = {
    'data': {
        'source': [
            'import time',
            'time.sleep(0.3)',
            '',
            'def f():',
            '    for _ in range(10):',
            '        pass',
            '    return',
            '',
            'f()',
            'f()',
            '',
            "print('here!')",
            '',
            'import script_threading',
            'script_threading.run()',
            '',
            '',
            'import script_asyncio',
            'script_asyncio.run()'
        ]
    }
}

snapshots['test_source[path] 1'] = {
    'data': {
        'source': [
            'from threading import Thread',
            '',
            'def run():',
            '    t1 = Thread(target=f1)',
            '    t1.start()',
            '    t2 = Thread(target=f2)',
            '    t2.start()',
            '    t2.join()',
            '    t1.join()',
            '    return',
            '',
            'def f1():',
            '    for _ in range(5):',
            '        pass',
            '    return',
            '',
            'def f2():',
            '    for _ in range(5):',
            '        pass',
            '    return',
            ''
        ]
    }
}

snapshots['test_source[string] 1'] = {
    'data': {
        'source': [
            'import time',
            'time.sleep(0.3)',
            '',
            'def f():',
            '    for _ in range(10):',
            '        pass',
            '    return',
            '',
            'f()',
            'f()',
            '',
            "print('here!')",
            '',
            'import script_threading',
            'script_threading.run()',
            '',
            '',
            'import script_asyncio',
            'script_asyncio.run()'
        ]
    }
}



import asyncio
from async_asgi_testclient import TestClient

import pytest

from nextlinegraphql import create_app

from .gql import (
    QUERY_GLOBAL_STATE,
    QUERY_SOURCE_LINE,
    SUBSCRIBE_GLOBAL_STATE,
    SUBSCRIBE_THREAD_TASK_IDS,
    SUBSCRIBE_THREAD_TASK_STATE,
    MUTATE_EXEC,
    MUTATE_SEND_PDB_COMMAND,
)


##__________________________________________________________________||
async def control_thread_task(client, thread_task_id):
    # print(f'control_thread_task({thread_task_id})')

    to_step = ["script_threading.run()", "script_asyncio.run()"]

    subscribe_thread_task_state = {
        "id": "1",
        "type": "start",
        "payload": {
            "variables": thread_task_id,
            "extensions": {},
            "operationName": None,
            "query": SUBSCRIBE_THREAD_TASK_STATE,
        },
    }

    query_source_line = {"query": QUERY_SOURCE_LINE}

    mutate_send_pdb_command = {"query": MUTATE_SEND_PDB_COMMAND}

    headers = {"Content-Type:": "application/json"}

    async with client.websocket_connect("/") as ws:
        await ws.send_json(subscribe_thread_task_state)
        while True:
            resp_json = await ws.receive_json()
            if resp_json["type"] == "complete":
                break
            assert "errors" not in resp_json["payload"]
            state = resp_json["payload"]["data"]["threadTaskState"]
            # print(state)
            if state["prompting"]:
                command = "next"
                if state["traceEvent"] == "line":
                    query_source_line["variables"] = {
                        "lineNo": state["lineNo"],
                        "fileName": state["fileName"],
                    }
                    resp = await client.post(
                        "/", json=query_source_line, headers=headers
                    )
                    source_line = resp.json()["data"]["sourceLine"]

                    # print(source_line)
                    # print(source_line in to_step)
                    if source_line in to_step:
                        command = "step"

                mutate_send_pdb_command["variables"] = {
                    **thread_task_id,
                    "command": command,
                }
                resp = await client.post(
                    "/", json=mutate_send_pdb_command, headers=headers
                )


async def control_execution(client):

    subscribe_thread_task_ids = {
        "id": "1",
        "type": "start",
        "payload": {
            "variables": {},
            "extensions": {},
            "operationName": None,
            "query": SUBSCRIBE_THREAD_TASK_IDS,
        },
    }

    async with client.websocket_connect("/") as ws:
        await ws.send_json(subscribe_thread_task_ids)
        prev_ids = set()
        tasks_control_thread_task = set()
        task_receive_json = asyncio.create_task(ws.receive_json())
        while True:
            aws = {task_receive_json, *tasks_control_thread_task}
            done, pending = await asyncio.wait(
                aws, return_when=asyncio.FIRST_COMPLETED
            )
            results = [
                t.result() for t in tasks_control_thread_task & done
            ]  # raise exception
            tasks_control_thread_task = tasks_control_thread_task & pending
            if task_receive_json not in done:
                continue
            resp_json = task_receive_json.result()
            task_receive_json = asyncio.create_task(ws.receive_json())
            print(resp_json)
            if resp_json["type"] == "complete":
                break
            ids = resp_json["payload"]["data"][
                "threadTaskIds"
            ]  # a list of dicts
            if not ids:
                break
            ids = {
                tuple(id_.items()) for id_ in ids
            }  # a set of tuples. note: a dict is unhashable
            new_ids = ids - prev_ids
            for id_ in new_ids:
                task = asyncio.create_task(
                    control_thread_task(client, dict(id_))
                )
                tasks_control_thread_task.add(task)
            prev_ids = ids


async def monitor_global_state(client):

    subscribe_global_state = {
        "id": "1",
        "type": "start",
        "payload": {
            "variables": {},
            "extensions": {},
            "operationName": None,
            "query": SUBSCRIBE_GLOBAL_STATE,
        },
    }

    async with client.websocket_connect("/") as ws:
        await ws.send_json(subscribe_global_state)
        while True:
            resp_json = await ws.receive_json()
            if resp_json["type"] == "complete":
                break
            # print(resp_json['payload']['data']['globalState'])
            if resp_json["payload"]["data"]["globalState"] == "finished":
                break


@pytest.mark.asyncio
async def test_run(snapshot):

    query_global_state = {"query": QUERY_GLOBAL_STATE}

    mutate_exec = {"query": MUTATE_EXEC}

    headers = {"Content-Type:": "application/json"}

    async with TestClient(create_app()) as client:
        resp = await client.post("/", json=query_global_state, headers=headers)
        assert "initialized" == resp.json()["data"]["globalState"]

        task_monitor_global_state = asyncio.create_task(
            monitor_global_state(client)
        )

        resp = await client.post("/", json=mutate_exec, headers=headers)
        assert resp.json()["data"]["exec"]

        task_control_execution = asyncio.create_task(control_execution(client))

        aws = {task_monitor_global_state, task_control_execution}
        while aws:
            done, pending = await asyncio.wait(
                aws, return_when=asyncio.FIRST_COMPLETED
            )
            results = [t.result() for t in done]  # re-raise exception
            aws = pending
            # break # to be removed

        resp = await client.post("/", json=query_global_state, headers=headers)
        assert "finished" == resp.json()["data"]["globalState"]


##__________________________________________________________________||



import asyncio

from nextline import Nextline

from .example_script import statement


nextline_holder = []


def get_nextline():
    if not nextline_holder:
        nextline_holder.append(Nextline(statement))
    return nextline_holder[0]


async def run_nextline():
    nextline = get_nextline()
    # print(nextline.global_state)
    if nextline.global_state == "initialized":
        nextline.run()
        asyncio.create_task(_wait(nextline))
    return


async def _wait(nextline):
    await nextline.finish()


def reset_nextline(statement=None):
    nextline = get_nextline()
    nextline.reset(statement=statement)


def pop_nextline():
    nextline = get_nextline()
    nextline_holder.clear()
    return nextline


async def close_nextline():
    nextline = pop_nextline()
    await nextline.close()



from ariadne import graphql

import pytest

from nextlinegraphql import schema


##__________________________________________________________________||
@pytest.mark.xfail()
@pytest.mark.asyncio
async def test_schema(snapshot):

    query = """
      {
        __schema {
          types {
            name
            description
            fields {
              name
              description
              type {
                name
              }
            }
            inputFields {
              name
              description
              defaultValue
            }
          }
        }
      }
    """

    data = {"query": query}
    success, response = await graphql(schema, data)
    assert success
    assert "errors" not in response
    snapshot.assert_match(response)


##__________________________________________________________________||



from setuptools import setup, find_packages
import versioneer

from pathlib import Path

here = Path(__file__).resolve().parent
long_description = here.joinpath("README.md").read_text()

setup(
    name="nextline-graphql",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="A GraphQL API for Nextline",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Simons Observatory",
    author_email="so_software@simonsobservatory.org",
    url="https://github.com/simonsobs/nextline",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    packages=find_packages(exclude=["docs", "tests"]),
    include_package_data=True,
    install_requires=[
        "nextline>=0.0.3",
        "ariadne>=0.12.0",
        "uvicorn>=0.12.2",
        "websockets>=8.1",
        "janus>=0.6",
        "SQLAlchemy>=1.4",
    ],
    extras_require={
        "tests": [
            "pytest>=5.4",
            "pytest-cov>=2.8",
            "pytest-asyncio>=0.14",
            "snapshottest>0.5",
            "async-asgi-testclient>=1.4.6",
        ],
        "dev": [
            "uvicorn",
            "twine",
            "flake8",
            "black",
        ],
    },
)



from pathlib import Path

import pytest

from async_asgi_testclient import TestClient

from nextlinegraphql import create_app

from ..gql import QUERY_SOURCE

##__________________________________________________________________||
THIS_DIR = Path(__file__).resolve().parent
PACKAGE_TOP = THIS_DIR.parent.parent.parent
SCRIPT_PATH = str(
    PACKAGE_TOP.joinpath(
        "nextlinegraphql", "nl", "example_script", "script_threading.py"
    )
)

params = [
    pytest.param(None, id="default"),
    pytest.param("<string>", id="string"),
    pytest.param(SCRIPT_PATH, id="path"),
]


@pytest.mark.parametrize("file_name", params)
@pytest.mark.asyncio
async def test_source(snapshot, file_name):

    data = {"query": QUERY_SOURCE}

    if file_name:
        data["variables"] = {"fileName": file_name}

    headers = {"Content-Type:": "application/json"}

    async with TestClient(create_app()) as client:
        resp = await client.post("/", json=data, headers=headers)
        assert resp.status_code == 200
        snapshot.assert_match(resp.json())


##__________________________________________________________________||



from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from . import models

SQLALCHEMY_DATABASE_URL = "sqlite:///:memory:"


class Db:
    def __init__(self):
        self.engine = create_engine(
            SQLALCHEMY_DATABASE_URL, connect_args={"check_same_thread": False}
        )
        self.Session = sessionmaker(
            autocommit=False, autoflush=False, bind=self.engine
        )
        self.models = models
        self.models.Base.metadata.create_all(bind=self.engine)



import pytest

from async_asgi_testclient import TestClient

from nextlinegraphql import create_app

from ..gql import QUERY_SOURCE, MUTATION_RESET

##__________________________________________________________________||
SOURCE_ONE = """
import time
time.sleep(0.1)
""".strip()

##__________________________________________________________________||
params = [
    pytest.param(None, id="no-statement"),
    pytest.param(SOURCE_ONE, id="statement"),
]


@pytest.mark.parametrize("statement", params)
@pytest.mark.asyncio
async def test_reset(snapshot, statement):

    headers = {"Content-Type:": "application/json"}

    data = {"query": MUTATION_RESET}
    if statement:
        data["variables"] = {"statement": statement}

    async with TestClient(create_app()) as client:
        resp = await client.post("/", json=data, headers=headers)
        assert resp.status_code == 200
        assert {"data": {"reset": True}} == resp.json()

    data = {"query": QUERY_SOURCE}

    async with TestClient(create_app()) as client:
        resp = await client.post("/", json=data, headers=headers)
        assert resp.status_code == 200
        snapshot.assert_match(resp.json())


##__________________________________________________________________||



"""
uvicorn --factory --reload --reload-dir nextline-graphql --reload-dir nextline nextlinegraphql:create_app
"""

__all__ = ["create_app"]

from .factory import create_app

##__________________________________________________________________||
# remove args so that they won't be processed in the executing script
# TODO This should be properly handled and tested
import sys  # noqa: E402

sys.argv[1:] = []

##__________________________________________________________________||
from ._version import get_versions  # noqa: E402

__version__ = get_versions()["version"]
"""str: version

The version string, e.g., "0.1.2", "0.1.2+83.ga093a20.dirty".
generated from git tags by versioneer.

Versioneer: https://github.com/warner/python-versioneer

"""

del get_versions

##__________________________________________________________________||



import sys
from ariadne import graphql, subscribe

import pytest
from unittest.mock import AsyncMock, Mock

from nextlinegraphql.schema import schema


##__________________________________________________________________||
@pytest.fixture(autouse=True)
def mock_asyncio_sleep(monkeypatch):
    import asyncio

    y = Mock(wraps=asyncio)
    y.sleep = AsyncMock()
    module = sys.modules["nextlinegraphql.schema.bindables"]
    monkeypatch.setattr(module, "asyncio", y)
    yield y


##__________________________________________________________________||
@pytest.mark.asyncio
async def test_query():

    query = """
      { hello }
    """
    data = {"query": query}

    request = Mock()
    request.headers = {"user-agent": "Mozilla/5.0"}
    context_value = {"request": request}
    success, response = await graphql(
        schema, data, context_value=context_value
    )
    assert success
    assert "errors" not in response
    expect = {"data": {"hello": "Hello, Mozilla/5.0!"}}
    assert expect == response


##__________________________________________________________________||
@pytest.mark.asyncio
async def test_subscription():

    query = """
      subscription {
        counter
      }
    """

    data = {"query": query}
    success, result = await subscribe(schema, data)
    assert success

    response = await result.__anext__()
    expect = {"counter": 1}
    assert expect == response.data

    response = await result.__anext__()
    expect = {"counter": 2}
    assert expect == response.data

    response = await result.__anext__()
    expect = {"counter": 3}
    assert expect == response.data

    response = await result.__anext__()
    expect = {"counter": 4}
    assert expect == response.data

    response = await result.__anext__()
    expect = {"counter": 5}
    assert expect == response.data

    with pytest.raises(StopAsyncIteration):
        response = await result.__anext__()


##__________________________________________________________________||
